@@ -42,12 +42,14 @@
 
 #include "yelp-bookmarks.h"
 #include "yelp-db-pager.h"
+#include "yelp-db-print-pager.h"
 #include "yelp-error.h"
 #include "yelp-html.h"
 #include "yelp-pager.h"
 #include "yelp-settings.h"
 #include "yelp-toc-pager.h"
 #include "yelp-window.h"
+#include "yelp-print.h"
 
 #ifdef ENABLE_MAN
 #include "yelp-man-pager.h"
@@ -113,8 +115,6 @@ static void        window_handle_page             (YelpWindow        *window,
 static void        window_disconnect              (YelpWindow        *window);
 
 /** Window Callbacks **/
-static void        yelp_window_destroyed          (GtkWidget         *window,
-						   gpointer           user_data);
 static gboolean    window_configure_cb            (GtkWidget         *widget,
 						   GdkEventConfigure *event,
 						   gpointer           data);
@@ -137,6 +137,10 @@ static void        html_uri_selected_cb           (YelpHtml          *html,
 						   gchar             *uri,
 						   gboolean           handled,
 						   gpointer           user_data);
+static gboolean    html_frame_selected_cb         (YelpHtml          *html,
+						   gchar             *uri,
+						   gboolean           handled,
+						   gpointer           user_data);
 static void        html_title_changed_cb          (YelpHtml          *html,
 						   gchar             *title,
 						   gpointer           user_data);
@@ -160,6 +164,8 @@ static void    window_add_widget        (GtkUIManager *ui_manager,
 					 GtkWidget    *vbox);
 static void    window_new_window_cb     (GtkAction *action, YelpWindow *window);
 static void    window_about_document_cb (GtkAction *action, YelpWindow *window);
+static void    window_print_document_cb (GtkAction *action, YelpWindow *window);
+static void    window_print_page_cb    (GtkAction *action, YelpWindow *window);
 static void    window_open_location_cb  (GtkAction *action, YelpWindow *window);
 static void    window_close_window_cb   (GtkAction *action, YelpWindow *window);
 static void    window_copy_cb           (GtkAction *action, YelpWindow *window);
@@ -213,6 +219,8 @@ static void        window_find_clicked_cb         (GtkWidget         *button,
 						   YelpWindow        *window);
 static void        window_find_next_cb            (GtkAction *action, 
 						   YelpWindow *window);
+static void        window_find_previous_cb        (GtkAction *action, 
+						   YelpWindow *window);
 static gboolean    tree_model_iter_following      (GtkTreeModel      *model,
 						   GtkTreeIter       *iter);
 
@@ -240,6 +248,8 @@ struct _YelpWindowPriv {
     GtkToolItem    *find_prev;
     GtkToolItem    *find_next;
     gchar          *find_string;
+    GtkAction      *find_next_menu;
+    GtkAction      *find_prev_menu;
 
     /* Open Location */
     GtkWidget      *location_dialog;
@@ -309,6 +319,16 @@ static const GtkActionEntry entries[] = {
       "<Control>N",
       NULL,
       G_CALLBACK (window_new_window_cb) },
+    { "PrintDocument", NULL,
+      N_("Print This Document"),
+      NULL,
+      NULL,
+      G_CALLBACK (window_print_document_cb) },
+    { "PrintPage", NULL,
+      N_("Print This Page"),
+      NULL,
+      NULL,
+      G_CALLBACK (window_print_page_cb) },
     { "AboutDocument", NULL,
       N_("About This Document"),
       NULL,
@@ -341,10 +361,15 @@ static const GtkActionEntry entries[] = {
       "<Control>F",
       NULL,
       G_CALLBACK (window_find_cb) },
+    { "FindPrev", NULL,
+      N_("Find Pre_vious"),
+      "<Control><shift>G",
+      N_("Find previous occurrence of the word or phrase"),
+      G_CALLBACK (window_find_previous_cb) },
     { "FindNext", NULL,
       N_("Find Ne_xt"),
       "<Control>G",
-      NULL,
+      N_("Find next occurrence of the word or phrase"),
       G_CALLBACK (window_find_next_cb) },
     { "Preferences", GTK_STOCK_PREFERENCES,
       N_("_Preferences"),
@@ -477,6 +502,18 @@ window_init (YelpWindow *window)
 		      NULL);
 
     gtk_window_set_title (GTK_WINDOW (window), _("Help Browser"));
+   
+    window_populate (window);
+}
+
+static void
+window_dispose (GObject *object)
+{
+    YelpWindow *window = YELP_WINDOW (object);
+
+    window_disconnect (YELP_WINDOW (window));
+
+    parent_class->dispose (object);
 }
 
 static void
@@ -490,6 +527,10 @@ window_finalize (GObject *object)
 
     g_free (priv->find_string);
     
+    if (priv->current_doc)
+	    yelp_doc_info_unref (priv->current_doc);
+    g_free (priv->current_frag);
+
     /* FIXME there are many more things to free */
 
     parent_class->finalize (object);
@@ -502,6 +543,7 @@ window_class_init (YelpWindowClass *klass)
 
     parent_class = (GObjectClass *) g_type_class_peek_parent (klass);
 
+    object_class->dispose = window_dispose;
     object_class->finalize = window_finalize;
 
     signals[NEW_WINDOW_REQUESTED] =
@@ -628,7 +670,8 @@ history_pop_back (YelpWindow *window)
     GtkAction        *action;
 
     g_return_val_if_fail (YELP_IS_WINDOW (window), NULL);
-    g_return_val_if_fail (window->priv->history_back != NULL, NULL);
+    if (!window->priv->history_back)
+	return NULL;
 
     priv = window->priv;
 
@@ -652,7 +695,8 @@ history_pop_forward (YelpWindow *window)
     GtkAction        *action;
 
     g_return_val_if_fail (YELP_IS_WINDOW (window), NULL);
-    g_return_val_if_fail (window->priv->history_forward != NULL, NULL);
+    if (!window->priv->history_forward)
+	return NULL;
 
     priv = window->priv;
 
@@ -697,22 +741,7 @@ load_data_free (YelpLoadData *data)
 GtkWidget *
 yelp_window_new (GNode *doc_tree, GList *index)
 {
-    YelpWindow     *window;
-    YelpWindowPriv *priv;
-
-    window = g_object_new (YELP_TYPE_WINDOW, NULL);
-
-    priv   = window->priv;
-
-    d (g_print ("yelp_window_new\n"));
-
-    window_populate (window);
-
-    g_signal_connect (G_OBJECT (window), "destroy",
-		      G_CALLBACK (yelp_window_destroyed),
-		      NULL);
-
-    return GTK_WIDGET (window);
+    return GTK_WIDGET (g_object_new (YELP_TYPE_WINDOW, NULL));
 }
 
 void
@@ -721,13 +750,19 @@ yelp_window_load (YelpWindow *window, const gchar *uri)
     YelpWindowPriv *priv;
     YelpDocInfo    *doc_info;
     gchar          *frag_id;
-
+    GtkAction      *action;
+    gchar          *real_uri;
     g_return_if_fail (YELP_IS_WINDOW (window));
 
+    if (g_str_has_prefix (uri, "info:") && g_str_has_suffix (uri, "dir")) {
+	real_uri = g_strdup ("x-yelp-toc:#Info");
+    } else {
+	real_uri = g_strdup (uri);
+    }
     d (g_print ("yelp_window_load\n"));
     d (g_print ("  uri = \"%s\"\n", uri));
 
-    if (!uri) {
+    if (!real_uri) {
 	GError *error = NULL;
 	g_set_error (&error, YELP_ERROR, YELP_ERROR_NO_DOC,
 		     _("The Uniform Resource Identifier for the file is "
@@ -738,17 +773,17 @@ yelp_window_load (YelpWindow *window, const gchar *uri)
 
     priv = window->priv;
 
-    doc_info = yelp_doc_info_get (uri);
+    doc_info = yelp_doc_info_get (real_uri, FALSE);
     if (!doc_info) {
 	GError *error = NULL;
 	g_set_error (&error, YELP_ERROR, YELP_ERROR_NO_DOC,
 		     _("The Uniform Resource Identifier ‘%s’ is invalid "
 		       "or does not point to an actual file."),
-		     uri);
+		     real_uri);
 	window_error (window, error, FALSE);
 	return;
     }
-    frag_id  = yelp_uri_get_fragment (uri);
+    frag_id  = yelp_uri_get_fragment (real_uri);
 
     if (priv->current_doc && yelp_doc_info_equal (priv->current_doc, doc_info)) {
 	if (priv->current_frag) {
@@ -768,6 +803,11 @@ yelp_window_load (YelpWindow *window, const gchar *uri)
     if (priv->current_frag)
 	g_free (priv->current_frag);
 
+    action = gtk_action_group_get_action (priv->action_group, 
+						  "PrintDocument");
+    g_object_set (G_OBJECT (action), "sensitive", FALSE, NULL);
+
+
     priv->current_doc  = yelp_doc_info_ref (doc_info);
     priv->current_frag = g_strdup (frag_id);
 
@@ -776,6 +816,7 @@ yelp_window_load (YelpWindow *window, const gchar *uri)
 
     if (priv->current_frag != frag_id)
 	g_free (frag_id);
+    g_free (real_uri);
 }
 
 YelpDocInfo *
@@ -799,9 +840,8 @@ window_do_load (YelpWindow  *window,
 		YelpDocInfo *doc_info,
 		gchar       *frag_id)
 {
-    YelpWindowPriv *priv;
+    GtkAction      *action;
     GError         *error = NULL;
-    gboolean        handled = FALSE;
     gchar          *uri;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
@@ -809,12 +849,17 @@ window_do_load (YelpWindow  *window,
 
     d (g_print ("window_do_laod\n"));
 
-    priv = window->priv;
-
+    action = gtk_action_group_get_action (window->priv->action_group, 
+					  "PrintDocument");
+    g_object_set (G_OBJECT (action),
+		  "sensitive",
+		  FALSE,
+		  NULL);
+    
     switch (yelp_doc_info_get_type (doc_info)) {
     case YELP_DOC_TYPE_MAN:
 #ifdef ENABLE_MAN
-	handled = window_do_load_pager (window, doc_info, frag_id);
+	window_do_load_pager (window, doc_info, frag_id);
 	break;
 #else
 	g_set_error (&error, YELP_ERROR, YELP_ERROR_FORMAT,
@@ -824,7 +869,7 @@ window_do_load (YelpWindow  *window,
 
     case YELP_DOC_TYPE_INFO:
 #ifdef ENABLE_INFO
-	handled = window_do_load_pager (window, doc_info, frag_id);
+	window_do_load_pager (window, doc_info, frag_id);
 	break;
 #else
 	g_set_error (&error, YELP_ERROR, YELP_ERROR_FORMAT,
@@ -833,12 +878,16 @@ window_do_load (YelpWindow  *window,
 #endif
 
     case YELP_DOC_TYPE_DOCBOOK_XML:
+	g_object_set (G_OBJECT (action),
+		      "sensitive",
+		      TRUE,
+		      NULL);
     case YELP_DOC_TYPE_TOC:
-	handled = window_do_load_pager (window, doc_info, frag_id);
+	window_do_load_pager (window, doc_info, frag_id);
 	break;
     case YELP_DOC_TYPE_SEARCH:
 #ifdef ENABLE_SEARCH
-	handled = window_do_load_pager (window, doc_info, frag_id);
+	window_do_load_pager (window, doc_info, frag_id);
 	break;
 #else
 	g_set_error (&error, YELP_ERROR, YELP_ERROR_FORMAT,
@@ -852,7 +901,7 @@ window_do_load (YelpWindow  *window,
 	break;
     case YELP_DOC_TYPE_HTML:
     case YELP_DOC_TYPE_XHTML:
-	handled = window_do_load_html (window, doc_info, frag_id);
+	window_do_load_html (window, doc_info, frag_id);
 	break;
     case YELP_DOC_TYPE_EXTERNAL:
 	history_step_back (window);
@@ -973,13 +1022,13 @@ window_populate (YelpWindow *window)
     
 #ifdef ENABLE_SEARCH
     action =  gtk_entry_action_new ("Search",
-				    _("Search"),
+				    _("_Search:"),
 				    _("Search for other documentation"),
 				    NULL);
     g_signal_connect (G_OBJECT (action), "activate",
 		      G_CALLBACK (search_activated), window);
-#endif
     gtk_action_group_add_action (priv->action_group, action);
+#endif
 
     priv->ui_manager = gtk_ui_manager_new ();
     gtk_ui_manager_insert_action_group (priv->ui_manager, priv->action_group, 0);
@@ -1015,17 +1064,21 @@ window_populate (YelpWindow *window)
     action = gtk_action_group_get_action (priv->action_group, "GoForward");
     if (action)
 	g_object_set (G_OBJECT (action), "sensitive", FALSE, NULL);
+
     priv->popup = gtk_ui_manager_get_widget(priv->ui_manager, "ui/main_popup");
     priv->maillink = gtk_ui_manager_get_widget(priv->ui_manager, "ui/mail_popup");
     priv->merge_id = gtk_ui_manager_new_merge_id (priv->ui_manager);
 
+    priv->find_next_menu = gtk_action_group_get_action (priv->action_group, 
+							"FindNext");
+    priv->find_prev_menu = gtk_action_group_get_action (priv->action_group, 
+						      "FindPrev");
+
     priv->pane = gtk_hpaned_new ();
-    gtk_widget_ref (priv->pane);
     /* We should probably remember the last position and set to that. */
     gtk_paned_set_position (GTK_PANED (priv->pane), 180);
 
     priv->side_sw = gtk_scrolled_window_new (NULL, NULL);
-    gtk_widget_ref (priv->side_sw);
     gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (priv->side_sw),
 				    GTK_POLICY_AUTOMATIC,
 				    GTK_POLICY_AUTOMATIC);
@@ -1076,6 +1129,10 @@ window_populate (YelpWindow *window)
 		      "uri_selected",
 		      G_CALLBACK (html_uri_selected_cb),
 		      window);
+    g_signal_connect (priv->html_view,
+		      "frame_selected",
+		      G_CALLBACK (html_frame_selected_cb),
+		      window);
     g_signal_connect (priv->html_view,
 		      "title_changed",
 		      G_CALLBACK (html_title_changed_cb),
@@ -1098,6 +1155,9 @@ window_populate (YelpWindow *window)
     gtk_widget_show (priv->html_pane);
     gtk_widget_show (priv->pane);
     gtk_widget_show (priv->main_box);
+    gtk_container_set_focus_child (GTK_CONTAINER (window),
+				   GTK_WIDGET (priv->html_view));
+
 }
 
 static void
@@ -1112,7 +1172,7 @@ window_populate_find (YelpWindow *window, GtkWidget *find_bar)
     g_return_if_fail (GTK_IS_TOOLBAR (find_bar));
 
     box = gtk_hbox_new (FALSE, 0);
-    label = gtk_label_new_with_mnemonic (_("Find:"));
+    label = gtk_label_new_with_mnemonic (_("Fin_d:"));
     gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 6);
 
     priv->find_entry = gtk_entry_new ();
@@ -1279,7 +1339,6 @@ window_do_load_pager (YelpWindow  *window,
 	    window_error (window, error, TRUE);
 	handled = FALSE;
 	goto done;
-	break;
     case YELP_PAGER_STATE_RUNNING:
     case YELP_PAGER_STATE_FINISHED:
 	/* Check if the page exists */
@@ -1427,6 +1486,13 @@ window_do_load_html (YelpWindow    *window,
 
     while ((result = gnome_vfs_read
 	    (handle, buffer, BUFFER_SIZE, &n)) == GNOME_VFS_OK) {
+	gchar *tmp;
+
+	tmp = g_utf8_strup (buffer, n);
+	if (strstr (tmp, "<FRAMESET"))
+	    yelp_html_frames (priv->html_view, TRUE);
+	g_free (tmp);
+
 	yelp_html_write (priv->html_view, buffer, n);
     }
 
@@ -1608,17 +1674,20 @@ window_disconnect (YelpWindow *window)
     g_return_if_fail (YELP_IS_WINDOW (window));
 
     priv = window->priv;
-    pager = yelp_doc_info_get_pager (priv->current_doc);
+    
+    if (priv && priv->current_doc) {
+	pager = yelp_doc_info_get_pager (priv->current_doc);
+    }
 
     if (GTK_WIDGET (window)->window)
 	gdk_window_set_cursor (GTK_WIDGET (window)->window, NULL);
 
-    if (window->priv->toc_pause > 0) {
-	window->priv->toc_pause--;
+    if (priv && priv->toc_pause > 0) {
+	priv->toc_pause--;
 	yelp_toc_pager_unpause (yelp_toc_pager_get ());
     }
 
-    if (window->priv->current_doc) {
+    if (priv && priv->current_doc) {
 	if (priv->start_handler) {
 	    g_signal_handler_disconnect (pager,
 					 priv->start_handler);
@@ -1645,7 +1714,7 @@ window_disconnect (YelpWindow *window)
 	    priv->finish_handler = 0;
 	}
     }
-    if (priv->idle_write) {
+    if (priv && priv->idle_write) {
 	gtk_idle_remove (priv->idle_write);
 	priv->idle_write = 0;
     }
@@ -1653,26 +1722,6 @@ window_disconnect (YelpWindow *window)
 
 /** Window Callbacks **********************************************************/
 
-static void
-yelp_window_destroyed (GtkWidget *window,
-		       gpointer   user_data)
-{
-    YelpWindowPriv *priv;
-
-    g_return_if_fail (YELP_IS_WINDOW (window));
-
-    priv = YELP_WINDOW(window)->priv;
-
-    window_disconnect (YELP_WINDOW (window));
-
-    if (priv->current_doc)
-	yelp_doc_info_unref (priv->current_doc);
-    g_free (priv->current_frag);
-
-    g_object_unref (priv->pane);
-    g_object_unref (priv->side_sw);
-}
-
 static gboolean
 window_configure_cb (GtkWidget         *widget,
 		     GdkEventConfigure *event,
@@ -1807,6 +1856,30 @@ html_uri_selected_cb (YelpHtml  *html,
     }
 }
 
+static gboolean
+html_frame_selected_cb (YelpHtml *html, gchar *uri, gboolean handled,
+			gpointer user_data)
+{
+    YelpWindow *window = YELP_WINDOW (user_data);
+    gboolean handle;
+    YelpDocInfo *info = yelp_doc_info_get (uri, FALSE);
+    switch (yelp_doc_info_get_type (info)) {
+    case YELP_DOC_TYPE_HTML:
+    case YELP_DOC_TYPE_XHTML:
+	handle = TRUE;
+	break;
+    default:
+	handle = FALSE;
+	break;
+    }
+    if (handle) {
+	return FALSE;
+    } else {
+	yelp_window_load (window, uri);
+    }
+    return TRUE;
+}
+
 static void
 html_title_changed_cb (YelpHtml  *html,
 		       gchar     *title,
@@ -1933,6 +2006,254 @@ window_new_window_cb (GtkAction *action, YelpWindow *window)
     g_signal_emit (window, signals[NEW_WINDOW_REQUESTED], 0, NULL);
 }
 
+typedef struct {
+    gulong page_handler;
+    gulong error_handler;
+    gulong cancel_handler;
+    gulong finish_handler;
+    YelpPager *pager;
+    YelpWindow *window;
+} PrintStruct;
+
+static void
+print_disconnect (PrintStruct *data)
+{
+    d(g_print ("print disconnect\n"));
+    if (data->page_handler) {
+	g_signal_handler_disconnect (data->pager,
+				     data->page_handler);
+	data->page_handler = 0;
+    }
+    if (data->error_handler) {
+	g_signal_handler_disconnect (data->pager,
+				     data->error_handler);
+	data->error_handler = 0;
+    }
+    if (data->cancel_handler) {
+	g_signal_handler_disconnect (data->pager,
+				     data->cancel_handler);
+	data->cancel_handler = 0;
+    }
+    if (data->finish_handler) {
+	g_signal_handler_disconnect (data->pager,
+				     data->finish_handler);
+	data->finish_handler = 0;
+    }
+    g_free (data);
+}
+
+static void
+print_pager_page_cb (YelpPager *pager,
+		     gchar     *page_id,
+		     gpointer   user_data)
+{
+    PrintStruct *data = user_data;
+    YelpPage    *page;
+    d (g_print ("print_pager_page_cb\n"));
+    d (g_print ("  page_id=\"%s\"\n", page_id));
+
+    page = (YelpPage *) yelp_pager_get_page (pager, page_id);
+
+    if (page) {
+	YelpHtml *html;
+	GtkWidget *gtk_window;
+	int length, offset;
+	char *uri;
+	GtkWidget *vbox = gtk_vbox_new (FALSE, FALSE);
+	d(g_print (page->contents));
+
+	gtk_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+	html = yelp_html_new ();
+
+	gtk_container_add (GTK_CONTAINER (gtk_window), GTK_WIDGET (vbox));
+	gtk_box_pack_end (GTK_BOX (vbox), GTK_WIDGET (html), TRUE, TRUE, 0);
+
+	gtk_widget_show (gtk_window);
+	gtk_widget_show (GTK_WIDGET (html));
+	gtk_widget_show (vbox);
+	gtk_widget_hide (gtk_window);
+	uri = yelp_doc_info_get_uri (yelp_pager_get_doc_info (pager),
+				     page_id,
+				     YELP_URI_TYPE_FILE);
+
+	d (g_print ("  uri            = %s\n", uri));
+
+	yelp_html_set_base_uri (html, uri);
+	g_free (uri);
+
+	yelp_html_open_stream (html, "application/xhtml+xml");
+	for (length = strlen (page->contents), offset = 0; length > 0; length -= BUFFER_SIZE, offset += BUFFER_SIZE) {
+	    d(g_print ("data: %.*s\n", MIN (length, BUFFER_SIZE), page->contents + offset));
+	    yelp_html_write (html, page->contents + offset, MIN (length, BUFFER_SIZE));
+	}
+	yelp_html_close (html);
+
+	yelp_print_run (data->window, html, gtk_window, vbox);
+
+	print_disconnect (data);
+
+    }
+}
+
+static void
+print_pager_error_cb (YelpPager   *pager,
+		      gpointer     user_data)
+{
+    PrintStruct *data = user_data;
+    /*     GError *error = yelp_pager_get_error (pager);*/
+
+    d (g_print ("print_pager_error_cb\n"));
+
+    print_disconnect (data);
+}
+
+static void
+print_pager_cancel_cb (YelpPager   *pager,
+		       gpointer     user_data)
+{
+    PrintStruct *data = user_data;
+    d (g_print ("print_pager_cancel_cb\n"));
+
+    print_disconnect (data);
+}
+
+static void
+print_pager_finish_cb (YelpPager   *pager,
+		       gpointer     user_data)
+{
+    PrintStruct *data = user_data;
+
+    d (g_print ("print_pager_finish_cb\n"));
+
+    print_disconnect (data);
+}
+
+static void
+window_print_document_cb (GtkAction *action, YelpWindow *window)
+{
+    PrintStruct *data;
+    YelpPager *pager;
+
+    if (!window->priv->current_doc)
+	return;
+
+    pager = yelp_db_print_pager_new (window->priv->current_doc);
+
+    if (!pager) {
+	return;
+    }
+
+    data = g_new0 (PrintStruct, 1);
+    data->pager = pager;
+    data->window = window;
+
+    data->page_handler =
+	g_signal_connect (data->pager,
+			  "page",
+			  G_CALLBACK (print_pager_page_cb),
+			  data);
+    data->error_handler =
+	g_signal_connect (data->pager,
+			  "error",
+			  G_CALLBACK (print_pager_error_cb),
+			  data);
+    data->cancel_handler =
+	g_signal_connect (data->pager,
+			  "error",
+			  G_CALLBACK (print_pager_cancel_cb),
+			  data);
+    data->finish_handler =
+	g_signal_connect (data->pager,
+			  "finish",
+			  G_CALLBACK (print_pager_finish_cb),
+			  data);
+
+    /* handled = */ yelp_pager_start (data->pager);
+}
+
+static void
+window_print_page_cb (GtkAction *action, YelpWindow *window)
+{
+    GtkWidget *gtk_win;
+    YelpPager  *pager;
+    YelpPage   *page  = NULL;
+    YelpHtml *html;
+    int length, offset;
+    gchar *uri;
+    GtkWidget *vbox = gtk_vbox_new (FALSE, FALSE);
+
+    gtk_win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+    html = yelp_html_new ();
+    
+    gtk_container_add (GTK_CONTAINER (gtk_win), GTK_WIDGET (vbox));
+    gtk_box_pack_end (GTK_BOX (vbox), GTK_WIDGET (html), TRUE, TRUE, 0);
+    gtk_widget_show (gtk_win);
+    gtk_widget_show (vbox);
+    gtk_widget_show (GTK_WIDGET (html));
+    gtk_widget_hide (gtk_win);
+
+    pager = yelp_doc_info_get_pager (window->priv->current_doc);
+
+    uri = yelp_doc_info_get_uri (window->priv->current_doc, NULL, YELP_URI_TYPE_FILE);
+    
+    yelp_html_set_base_uri (html, uri);
+
+    if (pager) {
+	page = (YelpPage *) yelp_pager_get_page (pager, window->priv->current_frag);
+	
+	yelp_html_open_stream (html, "application/xhtml+xml");
+	for (length = strlen (page->contents), offset = 0; length > 0; length -= BUFFER_SIZE, offset += BUFFER_SIZE) {
+	    d (g_print ("data: %.*s\n", MIN (length, BUFFER_SIZE), page->contents + offset));
+	    yelp_html_write (html, page->contents + offset, MIN (length, BUFFER_SIZE));
+	}
+	yelp_html_close (html);
+    } else { /*html file.  Dump file to window the easy way*/
+	GnomeVFSHandle  *handle;
+	GnomeVFSResult   result;
+	GnomeVFSFileSize n;
+	gchar            buffer[BUFFER_SIZE];	
+
+	result = gnome_vfs_open (&handle, uri, GNOME_VFS_OPEN_READ);
+	
+	if (result != GNOME_VFS_OK) {
+	    GError *error = NULL;
+	    g_set_error (&error, YELP_ERROR, YELP_ERROR_IO,
+			 _("The file ‘%s’ could not be read.  This file might "
+			   "be missing, or you might not have permissions to "
+			   "read it."),
+			 uri);
+	    window_error (window, error, TRUE);
+	    return;
+	}
+	/* Assuming the file exists.  If it doesn't how did we get this far?
+	 * There are more sinister forces at work...
+	 */
+
+	switch (yelp_doc_info_get_type (window->priv->current_doc)) {
+	case YELP_DOC_TYPE_HTML:
+	    yelp_html_open_stream (html, "text/html");
+	    break;
+	case YELP_DOC_TYPE_XHTML:
+	    yelp_html_open_stream (html, "application/xhtml+xml");
+	    break;
+	default:
+	    g_assert_not_reached ();
+	}
+	
+	while ((result = gnome_vfs_read
+		(handle, buffer, BUFFER_SIZE, &n)) == GNOME_VFS_OK) {
+	    yelp_html_write (html, buffer, n);
+	}
+	
+	yelp_html_close (html);
+
+	
+    
+    }
+    g_free (uri);
+    yelp_print_run (window, html, gtk_win, vbox);
+}
+
 static void
 window_about_document_cb (GtkAction *action, YelpWindow *window)
 {
@@ -2275,14 +2596,14 @@ static void window_copy_link_cb (GtkAction *action, YelpWindow *window)
 			    window->priv->uri,
 			    -1);
     g_free (window->priv->uri);
-};
+}
 
 static void
 window_open_link_cb (GtkAction *action, YelpWindow *window)
 {
     yelp_window_load (window, window->priv->uri);
     g_free (window->priv->uri);
-};
+}
 
 static void
 window_open_link_new_cb (GtkAction *action, YelpWindow *window)
@@ -2290,7 +2611,7 @@ window_open_link_new_cb (GtkAction *action, YelpWindow *window)
     g_signal_emit (window, signals[NEW_WINDOW_REQUESTED], 0,
 		   window->priv->uri);
     g_free (window->priv->uri);
-};
+}
 
 static void
 window_copy_mail_cb (GtkAction *action, YelpWindow *window)
@@ -2350,14 +2671,14 @@ location_response_cb (GtkDialog *dialog, gint id, YelpWindow *window)
     d (g_print ("  id = %i\n", id));
 
     if (id == GTK_RESPONSE_OK) {
-	const gchar *uri =
+	const gchar *uri = 
 	    gtk_entry_get_text (GTK_ENTRY (window->priv->location_entry));
+    
 	yelp_window_load (window, uri);
     }
 
     window->priv->location_dialog = NULL;
     window->priv->location_entry  = NULL;
-
     gtk_widget_destroy (GTK_WIDGET (dialog));
 }
 
@@ -2404,6 +2725,24 @@ window_find_again (YelpWindow *window, gboolean forward)
     return yelp_html_find_again (priv->html_view, forward);
 }
 
+static void
+window_find_previous_cb (GtkAction *action, YelpWindow *window)
+{
+    YelpWindowPriv * priv;
+
+    priv = window->priv;
+    if (!window_find_again (window, FALSE)) {
+	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), TRUE);
+	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), FALSE);
+	g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", TRUE, 
+		      NULL);
+	g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", FALSE, 
+		      NULL);
+    } else {
+	window_find_buttons_set_sensitive (window, TRUE);
+    }
+}
+
 static void
 window_find_next_cb (GtkAction *action, YelpWindow *window)
 {
@@ -2413,6 +2752,10 @@ window_find_next_cb (GtkAction *action, YelpWindow *window)
     if (!window_find_again (window, TRUE)) {
 	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), FALSE);
 	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), TRUE);
+	g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", FALSE, 
+		      NULL);
+	g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", TRUE, 
+		      NULL);
     } else {
 	window_find_buttons_set_sensitive (window, TRUE);
     }
@@ -2432,6 +2775,10 @@ window_find_clicked_cb (GtkWidget  *widget,
 	if (!window_find_again (window, TRUE)) {
 	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), FALSE);
 	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), TRUE);
+	    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", FALSE, 
+			  NULL);
+	    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", TRUE, 
+			  NULL);
 	} else {
 	    window_find_buttons_set_sensitive (window, TRUE);
 	}
@@ -2440,6 +2787,10 @@ window_find_clicked_cb (GtkWidget  *widget,
 	if (!window_find_again (window, FALSE)) {
 	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), TRUE);
 	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), FALSE);
+	    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", TRUE, 
+			  NULL);
+	    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", FALSE, 
+			  NULL);
 	} else {
 	    window_find_buttons_set_sensitive (window, TRUE);
 	}
@@ -2472,6 +2823,10 @@ window_find_entry_changed_cb (GtkEditable *editable,
 	if (found) {
 	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), TRUE);
 	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), FALSE);
+	    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", TRUE, 
+			  NULL);
+	    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", FALSE, 
+			  NULL);
 	} else {
 	    window_find_buttons_set_sensitive (window, TRUE);
 	}
@@ -2497,6 +2852,10 @@ window_find_entry_activate_cb (GtkEditable *editable,
     if (!window_find_again (window, TRUE)) {
 	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), FALSE);
 	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), TRUE);
+	g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", FALSE, 
+		      NULL);
+	g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", TRUE, 
+		      NULL);
     } else {
 	window_find_buttons_set_sensitive (window, TRUE);
     }
@@ -2536,6 +2895,10 @@ window_find_buttons_set_sensitive (YelpWindow  *window,
 
     gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), sensitive);
     gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), sensitive);
+    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", sensitive, 
+		  NULL);
+    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", sensitive, 
+		  NULL);
 }
 
 static gboolean
